#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <wiringPi.h>
#include <signal.h>
#include <string.h>
#include <time.h>
#include <libgen.h>

#include <unistd.h>
#include <sys/socket.h>
#include <bluetooth/bluetooth.h>
#include <bluetooth/rfcomm.h>

#include <errno.h>

#include "IrTimer.h"



//************************************************
//***** コンパイル
//sudo gcc IrTimer.c -lm -o IrTimer -lwiringPi



//************************************************
//*** 定数宣言
// 長押し判定の回数(bit数で指定)
//const unsigned long PUSH_LENGTH = 0x01FFFFFF;	// 25bit * 200ms = 5秒
const unsigned long PUSH_LENGTH = 0x00007FFF;	// 15bit * 200ms = 3秒
// インターバルタイマーの割り込み周期(ms)
const int INTARVAL = 200;

// コマンドID : 新フォーマット方式
const uint16_t CMD_ID_NEW_FORMAT		= 0x1000;
// コマンドID : 実行ボタン押下
const uint16_t CMD_ID_ACCEPT			= 0x1001;
// コマンドID : キャンセルボタン押下
const uint16_t CMD_ID_CANCEL			= 0x1002;
// コマンドID : コマンド名取得ボタン
const uint16_t CMD_ID_GET_CMD_NAME		= 0x1003;

// コマンドID : レスポンス
const uint16_t CMD_ID_RESPONSE			= 0x2000;

// レスポンス : 正常終了
const uint8_t RES_SUCCESS				= 0x00;
// レスポンス : 異常終了
const uint8_t RES_FAILURE				= 0x01;
// レスポンス : 実行中のため失敗
const uint8_t RES_FAILURE_FOR_RUNNING	= 0x02;
// レスポンス : 設定数が0のため失敗
const uint8_t RES_FAILURE_SETTING_ZERO	= 0x03;



//************************************************
//*** グローバル変数宣言
// 電源長押し判定用
unsigned long power_button;

// 実行中LED制御用
int running;

// 現在のステート/モード
STATE state;

// カウントダウンタイマーの設定値
unsigned int set_time;

// Bluetooth
int soc, client;

// インターバルタイマーによる割込みのマスクフラグ
int flag_it_mask;

// 時間設定構造体
unsigned int set_time_count;							// 時間設定構造体にセットされた有効なデータ数(1～)
unsigned int set_time_index;							// 時間設定構造体のindex(0～)
set_time_struct set_time_[SET_TIME_COUNT_MAX];			// 時間設定構造体の配列

// 設定
// 終了後の動作
COMPLETE_AFTER_SHUTDOWN_KIND complete_after_shutdown;



//************************************************
//*** メイン関数
//int main(void)
int main(int argc, char *argv[])
{
	//***** setitimer初期化
	struct sigaction action;
	memset(&action, 0, sizeof(action));

	// set signal alarm handler
	action.sa_handler	= interrupt_signal_alarm_handler;
	action.sa_flags		= SA_RESTART;
	sigemptyset(&action.sa_mask);
	if(sigaction(SIGALRM, &action, NULL) < 0)
	{
		printf("signal alarmの設定に失敗しました\n");
		exit(1);
	}

	// set intarval timer
	struct itimerval timer;
	timer.it_value.tv_sec		= 0;
	timer.it_value.tv_usec		= INTARVAL * 1000;
	timer.it_interval.tv_sec	= 0;
	timer.it_interval.tv_usec	= INTARVAL * 1000;
	if(setitimer(ITIMER_REAL, &timer, NULL) < 0)
	{
		printf("intarval timerの設定に失敗しました\n");
		exit(1);
	}



	//***** GPIO初期化
	// WiringPiの初期化
	if(wiringPiSetupGpio() == -1)
	{
		printf("WiringPiの初期化に失敗しました\n");
		exit(1);
	}

	// Set GPIO(LED) pin to output mode
	pinMode(LED_GREEN,  OUTPUT);
	pinMode(LED_ORANGE, OUTPUT);
	pinMode(LED_RED,    OUTPUT);

	// Set GPIO(SW) pin to input mode
	pinMode(SW_SHUTDOWN, INPUT);

	// キー割り込み設定
	wiringPiISR(SW_CANCEL, INT_EDGE_FALLING, interrupt_key_cancel_handler);
	wiringPiISR(SW_RECORD, INT_EDGE_FALLING, interrupt_key_record_handler);



	//***** Linuxのsignalキャッチ
	signal(SIGINT, signal_callback_handler);
    printf("press ^C to exit program ...\n");



	//***** その他初期化
	power_button	= 0;
	running			= 0;
	set_time_index	= 0;
	state			= STATE_IDLE;
	flag_it_mask	= false;
	complete_after_shutdown	= CAS_KEEP;

	// 起動を示すLED点灯
	digitalWrite(LED_GREEN, 1);



	//***** Bluetoothでの受信処理
	if(bluetooth_receive() == -1)
	{
		printf("Bluetooth処理異常終了\n");
	}
	else
	{
		printf("Bluetooth処理終了\n");		// ないハズ
	}



	// 終了時はLEDを消す
	digitalWrite(LED_GREEN,  OFF);
	digitalWrite(LED_ORANGE, OFF);
	digitalWrite(LED_RED,    OFF);

}



//************************************************
//*** 割り込みコールバック関数
void signal_callback_handler(int signum)
{
    digitalWrite(LED_GREEN, OFF);
    set_idle_state();

    close(client);
    close(soc);

    printf("\nProgram exit\n");
    exit(0);
}



//************************************************
//*** インターバルタイマーによる定期割り込み
void interrupt_signal_alarm_handler(int signum)
{
	// memo:
	// 毎回setitimer()にてインターバルタイマーを解除/有効にすると、
	// setitimer()による処理時間が長いため、flag処理で代用する。
	if(flag_it_mask == false)
	{
		// 長押し判定
		power_button = (power_button << 1) | !digitalRead(SW_SHUTDOWN);
		if((power_button & PUSH_LENGTH) == PUSH_LENGTH)
		{
			power_button = 0;
			shutdown_process();
		}



		// タイマー処理判定
		if(state != STATE_IDLE)
		{
			// 実行中を示すLEDの制御
			running++;
			if(running == 5)
			{
				digitalWrite(LED_ORANGE, OFF);
			}
			else if(running == 10)
			{
				digitalWrite(LED_ORANGE, ON);
				running = 0;
			}

			timer_end_judge();
		}
	}
}



//************************************************
//*** キー割り込み(キャンセル)
void interrupt_key_cancel_handler(void)
{
	if(state != STATE_IDLE)
	{
		set_idle_state();

		printf("キャンセルしました\n");
	}
}



//************************************************
//*** キー割り込み
void interrupt_key_record_handler(void)
{
	printf("interrupt_key_record_handler\n");
}



//************************************************
//*** タイマー動作を開始する
void set_start_state()
{
	unsigned int cur_time;

	// 実行中を示すLEDをon
	digitalWrite(LED_ORANGE, ON);
	running = 0;

	if(set_time_count > 0)
	{
		if(set_time_[set_time_index].kind == KIND_COUNTDOWN)
		{
			// 現在時刻を分単位で取得する
			get_cur_time(&cur_time);

			set_time = ((unsigned int)set_time_[set_time_index].hh * 60 * 60) + ((unsigned int)set_time_[set_time_index].mm * 60);
			set_time = cur_time + set_time;

			state = STATE_START_COUNTDOWN;

			// 直ちに実行かどうかの判定
			if((set_time_[set_time_index].hh == 0) && (set_time_[set_time_index].mm == 0))	// 0:00の場合は直ちに実行する
			{
				timer_end_judge();
			}
		}
		else if(set_time_[set_time_index].kind == KIND_TIMER)
		{
			// 現在時刻を分単位で取得する
			get_cur_time(&cur_time);

			set_time = ((unsigned int)set_time_[set_time_index].hh * 60 * 60) + ((unsigned int)set_time_[set_time_index].mm * 60);

			if(set_time < cur_time)		// 日跨ぎ処理
			{
				set_time = set_time + 86400;					// 86400 = 24 * 60 * 60 (24hの秒)
			}

			state = STATE_START_TIMER;

			// 直ちに実行かどうかの判定
			if(set_time_index > 1)
			{
				if( (set_time_[set_time_index].hh == set_time_[set_time_index-1].hh) &&	// 1つ前と同じ時刻なら直ちに実行
					(set_time_[set_time_index].mm == set_time_[set_time_index-1].mm))
				{
					timer_end_judge();
				}
			}
		}
	}
	else
	{
		set_idle_state();
	}
}



//************************************************
//*** カウントダウンが終了したかどうかの判定
void timer_end_judge()
{
	flag_it_mask = true;

	unsigned int cur_time;

	// 現在時刻を分単位で取得する
	get_cur_time(&cur_time);

	if((set_time - cur_time) <= 0)
	{
		if((set_time_count > 1) && ((set_time_index + 1) < set_time_count))		// 次のデータがある場合
		{
			send_ir_data();
			set_time_index++;
			set_start_state();
		}
		else																	// 次のデータは無し
		{
			send_ir_data();
			set_idle_state();

			if(complete_after_shutdown == CAS_SHUTDOWN)
			{
				usleep(500000);
				shutdown_process();
			}
		}
	}

	flag_it_mask = false;
}



//************************************************
//*** IDLEモードにする
void set_idle_state()
{
	state = STATE_IDLE;

	// 時間設定構造体
	set_time_index = 0;

	// 実行中を示すLEDをoff
	digitalWrite(LED_ORANGE, OFF);
}



//************************************************
//*** 現在の時間を取得し、秒単位にして返す
void get_cur_time(uint32_t* cur_time)
{
	time_t now = time(NULL);
	struct tm *pnow = localtime(&now);

	// 分単位にする
	*cur_time = (pnow->tm_hour      * 3600) +		// 数字演算のみでint16を超えるような演算をさせるとオーバーフローするので注意((24 * 60 * 60)の部分)
				(pnow->tm_min	     * 60)  +
				 pnow->tm_sec;
}



//************************************************
//*** 赤外線送信
void send_ir_data()
{
	time_t now = time(NULL);
	struct tm *pnow = localtime(&now);
	char cmdstr[512];

	sprintf(cmdstr, "irsend SEND_ONCE %s %s", set_time_[set_time_index].group, set_time_[set_time_index].cmd);
	printf("%2d:%02d:%02d : %s\n", pnow->tm_hour, pnow->tm_min, pnow->tm_sec, cmdstr);
	system(cmdstr);
	usleep(200000);		// memp: 連続で送信する場合を考慮して、間隔を開ける
}



//************************************************
//*** レスポンスの送信
void send_response(uint8_t res)
{
	uint8_t tx_buf[3];

	tx_buf[0] = (uint8_t) (0x00ff & CMD_ID_RESPONSE);
	tx_buf[1] = (uint8_t)((0xff00 & CMD_ID_RESPONSE) >> 8);
	tx_buf[2] = res;

	if(write(client, tx_buf, 3) == -1)
	{
		// memo: ここで失敗した場合、他に失敗通知する手段がないため、何もしない
	}
	else
	{
		// todo:
	}
}



//************************************************
//*** 受信処理
int data_receive(int fp, char *buf, int ln)
{
	char temp[ln];
	int i;
	int p;
	int ret;

	p = 0;
	i = 0;
	while(i<30)							// タイムアウトを「100ms x 30 = 3秒」とする
	{
		ret = read(fp, &temp[p], ln);
		if(ret == -1)
		{
			return -1;
		}
		else
		{
			if(ret > 0)
			{
				p = p + ret;
				if (p == ln)
				{
					memcpy(buf, temp, ln);
					return 0;
				}
				ln = ln - ret;
				i = 0;					// タイムアウトカントをクリアする
			}
			else
			{
				printf("data_receive() : ret=0\n");
				i++;
			}

			usleep(100000);				// 100ms
		}
	}

	printf("受信処理失敗(タイムアウト)\n");
	return -1;
}



//************************************************
//*** Bluetooth受信処理
int bluetooth_receive()
{
    struct sockaddr_rc loc_addr = { 0 }, rem_addr = { 0 };
    char buf[HEADER_SIZE * SET_TIME_COUNT_MAX] = { 0 };
    int bytes_read;
    socklen_t opt = sizeof(rem_addr);

    // allocate socket
    soc = socket(AF_BLUETOOTH, SOCK_STREAM, BTPROTO_RFCOMM);
    if(soc >= 0)			// -1 で失敗
    {
	    // bind socket to port 1 of the first available local bluetooth adapter
	    loc_addr.rc_family	= AF_BLUETOOTH;
	    loc_addr.rc_bdaddr	= *BDADDR_ANY;
	    loc_addr.rc_channel	= (uint8_t) 22;
	    bind(soc, (struct sockaddr *)&loc_addr, sizeof(loc_addr));

	    // put socket into listening mode
	    if(listen(soc, 1) == -1)
	    {
	    	printf("listenで失敗\n");
	    	close(soc);
	    	return -1;
	    }

	    // accept one connection
	accept:
	    client = accept(soc, (struct sockaddr *)&rem_addr, &opt);
	    if(client == -1)
	    {
	    	printf("acceptで失敗\n");
	    	close(soc);
	    	return -1;
	    }

	    ba2str( &rem_addr.rc_bdaddr, buf );
	    fprintf(stderr, "accepted connection from %s\n", buf);
	    memset(buf, 0, sizeof(buf));

	    // read data from the client
	    while(1)
	    {
	    	int flag;
	    	flag = true;

		    uint16_t cmd_id;
			if(data_receive(client, buf, 2) == -1)				// CMD ID取得
		    {
		    	printf("Bluetooth切断検知\n");					// BT切断時はここに来る
		    	goto accept;
		    }

			flag_it_mask = true;

		    cmd_id = ((uint16_t)buf[1] << 8) | buf[0];

			if(cmd_id == CMD_ID_NEW_FORMAT)
	 		{
	 			// ●新フォーマット方式
	 			int isRunning;
	 			isRunning = false;

				if(state == STATE_IDLE)
				{
		 			set_time_index = 0;
		 			set_time_count = 0;
				}
				else								// 実行中は受け付けない
				{
					isRunning = true;
				}

				uint8_t after_shutdown;
				if(data_receive(client, &after_shutdown, 1) != -1)			// 設定:終了後の動作
				{
					if(isRunning == false)
					{
						complete_after_shutdown = after_shutdown;
					}
				}
				else
				{
					if(isRunning == false)
					{
						complete_after_shutdown = CAS_KEEP;
					}
				}

				uint8_t cont;
				if(data_receive(client, &cont, 1) != -1)					// データ個数取得
	 			{
	 				int i;
	 				for(i=0; i<cont; i++)
	 				{
						if(data_receive(client, buf, HEADER_SIZE) != -1)	// データ取得
	 					{
	 						if(isRunning == false)							// 実行中は受け付けない
	 						{
								set_time_[i].kind	= buf[0];
								set_time_[i].hh		= buf[1];
								set_time_[i].mm		= buf[2];
								memcpy(set_time_[i].group, &buf[3], CMD_LENGTH);
								memcpy(set_time_[i].cmd, &buf[259], CMD_LENGTH);
								set_time_count++;
			 				}
			 			}
				 		else
				 		{
				 			flag = false;
				 			printf("データが足りない\n");
				 		}
	 				}

					if(isRunning == false)
					{
						if(flag)
						{
							printf("設定を受信しました\n");
							send_response(RES_SUCCESS);					// 成功

			 				set_start_state();
		 				}
		 				else
		 				{
							read(client, buf, sizeof(buf));				// ゴミデータを破棄
				 			printf("データが足りないため失敗\n");
				 			send_response(RES_FAILURE);
		 				}
					}
					else
					{
			 			printf("実行中のため実行できませんでした\n");
			 			send_response(RES_FAILURE_FOR_RUNNING);			// 実行中のためNG
					}
				}
			 	else
			 	{
			 		read(client, buf, sizeof(buf));				// ゴミデータを破棄
			 		printf("データが足りないため失敗\n");
			 		send_response(RES_FAILURE);					// データ個数取得に失敗
			 	}
	 		}
	 		else if(cmd_id == CMD_ID_ACCEPT)
	 		{
	 			// ●実行ボタン押下
	 			if(state == STATE_IDLE)						// 実行中は受け付けない
	 			{
	 				printf("実行しました\n");
	 				send_response(RES_SUCCESS);				// 成功
	 			}
	 			else
	 			{
	 				printf("既に実行中です\n");
	 				send_response(RES_FAILURE_FOR_RUNNING);	// 実行中のためNG
	 			}

	 			set_start_state();
	 		}
	 		else if(cmd_id == CMD_ID_CANCEL)
	 		{
	 			// ●キャンセルボタン押下
	 			printf("キャンセルしました\n");
	 			send_response(RES_SUCCESS);					// 成功

				set_idle_state();
	 		}
	 		else if(cmd_id == CMD_ID_GET_CMD_NAME)
	 		{
	 			// ●グループ名、コマンド名を取得する
				FILE *pp;

				char path[512];
				get_mypathname(path);						// memo: ここのみ、フルパスにする必要あり
				strcat(path, "/");
				strcat(path, FILE_NAME_GET_CMD_NAME);
				if((pp = popen(path, "r")) == NULL)
				{
					printf("コマンド取得に失敗しました(コマンド実行)\n");
					if(write(client, 0x00, 1) == -1)	// gcont=0 を送信して終了する
					{
						// 特に処理はしない
					}

					flag = false;
				}
				else
				{
					char stdout[CMD_LENGTH];
					while(fgets(stdout, sizeof(stdout), pp) != NULL)	// 標準出力を1行づつ読込む
					{
						stdout[strlen(stdout)-1] = '\0';				// 改行を削除

						char cmp[CMD_LENGTH];
						memset(cmp, 0, sizeof(cmp));					// 0埋めしている

						if(strncmp(stdout, "gcont=", 6) == 0)
						{
							strncpy(cmp, &stdout[6], (strlen(stdout)-6));
							uint8_t group_count[1];
							group_count[0]= atoi(cmp);
							if(write(client, group_count, 1) == -1)
							{
								printf("コマンド取得に失敗しました(グループ数取得)\n");
								flag = false;
								break;
							}
						}
						else if(strncmp(stdout, "group=", 6) == 0)
						{
							strncpy(cmp, &stdout[6], (strlen(stdout)-6));
							if(write(client, cmp, CMD_LENGTH) == -1)
							{
								printf("コマンド取得に失敗しました(グループ名取得)\n");
								flag = false;
								break;
							}
						}
						else if(strncmp(stdout, "ncont=", 6) == 0)
						{
							strncpy(cmp, &stdout[6], (strlen(stdout)-6));
							uint8_t cmd_count[1];
							cmd_count[0] = atoi(cmp);
							if(write(client, cmd_count, 1) == -1)
							{
								printf("コマンド取得に失敗しました(コマンド数取得)\n");
								flag = false;
								break;
							}
						}
						else if(strncmp(stdout, "name =", 6) == 0)
						{
							strncpy(cmp, &stdout[6], (strlen(stdout)-6));
							if(write(client, cmp, CMD_LENGTH) == -1)
							{
								printf("コマンド取得に失敗しました(コマンド名取得)\n");
								flag = false;
								break;
							}
						}
					}
				}

				pclose(pp);

	 			if(flag)
	 			{
	 				send_response(RES_SUCCESS);				// 成功
	 			}
	 			else
	 			{
	 				send_response(RES_FAILURE);				// 失敗
	 			}
	 		}
	 		else
	 		{
	 			state = STATE_ERROR;

				read(client, buf, sizeof(buf));

				printf("その他case\n");
	 			send_response(RES_FAILURE);
	 		}

			flag_it_mask = false;
		}

	    // close connection
	    close(client);
	    close(soc);
	}

	return -1;
}



//************************************************
//*** 自身のパスを取得する
void get_mypathname(char* path)
{
    static char buf[1024]={};

    readlink( "/proc/self/exe", buf, sizeof(buf)-1 );
    strcpy(path, dirname(buf));

    return;
}



//************************************************
//*** シャットダウン処理
void shutdown_process()
{
	// インターバルタイマーを停止する
	if(sigaction(SIGALRM, NULL, NULL) < 0)
	{
		perror("sigaction error");
	}

	printf("シャットダウンします。\n");

	// シャットダウンする事をLEDで知らせる
	int i;
	for(i=0; i<10; i++)
	{
		digitalWrite(LED_GREEN, OFF);
		usleep(150000);
		digitalWrite(LED_GREEN, ON);
		usleep(150000);
	}

    close(client);
    close(soc);

	system("shutdown -h now &");
	exit(0);
}



