/*
* Copyright (C) 2009 The Android Open Source Project
*
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*
* http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*/

using System.IO;
using System.Runtime.CompilerServices;
using Android.Bluetooth;
using Android.Content;
using Android.OS;
using Android.Util;
using Java.Lang;
using Java.Util;


namespace IrTimer
{
	/// <summary>
	/// This class does all the work for setting up and managing Bluetooth
	/// connections with other devices. It has a thread that listens for
	/// incoming connections, a thread for connecting with a device, and a
	/// thread for performing data transmissions when connected.
	/// </summary>
    class BluetoothService
	{
        #region 定数
		// Debugging
        private const string TAG = "BluetoothService";
		private const bool Debug = true;
	
		// Name for the SDP record when creating server socket
        private const string NAME = "IrTimer";
	
		// Unique UUID for this application
		//private static UUID MY_UUID = UUID.FromString ("fa87c0d0-afac-11de-8a39-0800200c9a66");
        // Unique UUID for SPP application
        private static UUID MY_UUID = UUID.FromString ("00001101-0000-1000-8000-00805F9B34FB");

        // Constants that indicate the current connection state
        // TODO: Convert to Enums
        public const int STATE_NONE = 0;       // we're doing nothing
        public const int STATE_LISTEN = 1;     // now listening for incoming connections
        public const int STATE_CONNECTING = 2; // now initiating an outgoing connection
        public const int STATE_CONNECTED = 3;  // now connected to a remote device

        // Message types sent from the BluetoothChatService Handler
        // TODO: Make into Enums
        public const int MESSAGE_STATE_CHANGE = 1;
        public const int MESSAGE_READ = 2;
        public const int MESSAGE_WRITE = 3;
        public const int MESSAGE_DEVICE_NAME = 4;
        public const int MESSAGE_TOAST = 5;


        /// <summary>受信タイムアウト(ms単位)</summary>
        private const int TIMEOUT_READ = 2400;
        /// <summary>受信タイムアウト(1byteあたり)(ms単位)</summary>
        private const int TIMEOUT_READ_BYTE = 200;
        #endregion //定数


        #region フィールド
		// Member fields
		protected BluetoothAdapter _adapter;
		protected Handler _handler;
		private AcceptThread acceptThread;
		protected ConnectThread connectThread;
		private ConnectedThread connectedThread;
		protected int _state;
        protected Context _context;
        #endregion //フィールド
	
		
        #region コンストラクタ
        /// <summary>
		/// Constructor. Prepares a new BluetoothChat session.
		/// </summary>
		/// <param name='context'>
		/// The UI Activity Context.
		/// </param>
		/// <param name='handler'>
		/// A Handler to send messages back to the UI Activity.
		/// </param>
        public BluetoothService (Context context, Handler handler)
		{
			_adapter = BluetoothAdapter.DefaultAdapter;
			_state = STATE_NONE;
			_handler = handler;
            _context = context;
		}
        #endregion //コンストラクタ
		

        #region SetState
		/// <summary>
		/// Set the current state of the chat connection.
		/// </summary>
		/// <param name='state'>
		/// An integer defining the current connection state.
		/// </param>
		[MethodImpl(MethodImplOptions.Synchronized)]
		private void SetState (int state)
		{
			if (Debug)
				Log.Debug (TAG, "setState() " + _state + " -> " + state);
			
			_state = state;
	
			// Give the new state to the Handler so the UI Activity can update
			_handler.ObtainMessage (MESSAGE_STATE_CHANGE, state, -1).SendToTarget ();
		}
        #endregion //SetState
		

        #region GetState
		/// <summary>
		/// Return the current connection state.
		/// </summary>
		[MethodImpl(MethodImplOptions.Synchronized)]
		public int GetState ()
		{
			return _state;
		}
        #endregion //GetState


        #region Start
		// Start the chat service. Specifically start AcceptThread to begin a
		// session in listening (server) mode. Called by the Activity onResume()
        // チャットサービスを開始します。
        // 具体的にリスニング（サーバー）モードでセッションを開始するAcceptThreadを開始します。
        // アクティビティのonResume（）によって呼び出される。
		[MethodImpl(MethodImplOptions.Synchronized)]
		public void Start ()
		{	
			if (Debug)
				Log.Debug (TAG, "start");
	
			// Cancel any thread attempting to make a connection
            // 接続を行おうとする任意のスレッドをキャンセルします。
			if (connectThread != null)
            {
				connectThread.Cancel ();
				connectThread = null;
			}
	
			// Cancel any thread currently running a connection
            // 現在の接続を実行している任意のスレッドをキャンセルします。
			if (connectedThread != null)
            {
				connectedThread.Cancel ();
				connectedThread = null;
			}
	
			// Start the thread to listen on a BluetoothServerSocket
            // BluetoothServerSocketをリッスンするためのスレッドを起動します。
			if (acceptThread == null)
            {
				acceptThread = new AcceptThread (this);
				acceptThread.Start ();
			}
			
			SetState (STATE_LISTEN);
		}
        #endregion //Start
		

        #region Connect
		/// <summary>
		/// Start the ConnectThread to initiate a connection to a remote device.
		/// </summary>
		/// <param name='device'>
		/// The BluetoothDevice to connect.
		/// </param>
		[MethodImpl(MethodImplOptions.Synchronized)]
		public void Connect (BluetoothDevice device)
		{
			if (Debug)
				Log.Debug (TAG, "connect to: " + device);
	
			// Cancel any thread attempting to make a connection
			if (_state == STATE_CONNECTING)
            {
				if (connectThread != null)
                {
					connectThread.Cancel ();
					connectThread = null;
				}
			}
	
			// Cancel any thread currently running a connection
			if (connectedThread != null)
            {
				connectedThread.Cancel ();
				connectedThread = null;
			}
	
			// Start the thread to connect with the given device
            // 指定されたデバイスとの接続にスレッドを起動します。
			connectThread = new ConnectThread (device, this);
			connectThread.Start ();
			
			SetState (STATE_CONNECTING);
		}
        #endregion //Connect


        #region Connected
		/// <summary>
		/// Start the ConnectedThread to begin managing a Bluetooth connection
		/// </summary>
		/// <param name='socket'>
		/// The BluetoothSocket on which the connection was made.
		/// </param>
		/// <param name='device'>
		/// The BluetoothDevice that has been connected.
		/// </param>
		[MethodImpl(MethodImplOptions.Synchronized)]
		public void Connected (BluetoothSocket socket, BluetoothDevice device)
		{
			if (Debug)
				Log.Debug (TAG, "connected");
	
			// Cancel the thread that completed the connection
            if (connectThread != null)
            {
                connectThread.Cancel();
                connectThread = null;
            }
	
			// Cancel any thread currently running a connection
            if (connectedThread != null)
            {
                connectedThread.Cancel();
                connectedThread = null;
            }
	
			// Cancel the accept thread because we only want to connect to one device
            if (acceptThread != null)
            {
                acceptThread.Cancel();
                acceptThread = null;
            }
			
			// Start the thread to manage the connection and perform transmissions
			connectedThread = new ConnectedThread (socket, this);
			connectedThread.Start ();
	
			// Send the name of the connected device back to the UI Activity
			var msg = _handler.ObtainMessage (MESSAGE_DEVICE_NAME);
			Bundle bundle = new Bundle ();
            bundle.PutString (MainActivity.DEVICE_NAME, device.Name);
			msg.Data = bundle;
			_handler.SendMessage (msg);
	
			SetState (STATE_CONNECTED);
		}
        #endregion //Connected


        #region Stop
		/// <summary>
		/// Stop all threads.
		/// </summary>
		[MethodImpl(MethodImplOptions.Synchronized)]
		public void Stop ()
		{
			if (Debug)
				Log.Debug (TAG, "stop");
	
            if (connectThread != null)
            {
                connectThread.Cancel();
                connectThread = null;
            }
	
            if (connectedThread != null)
            {
                connectedThread.Cancel();
                connectedThread = null;
            }
	
            if (acceptThread != null)
            {
                acceptThread.Cancel();
                acceptThread = null;
            }
			
			SetState (STATE_NONE);
		}
        #endregion //Stop


        #region Write
		/// <summary>
		/// Write to the ConnectedThread in an unsynchronized manner
		/// </summary>
		/// <param name='out'>
		/// The bytes to write.
		/// </param>
		public void Write (byte[] @out)
		{
			// Create temporary object
			ConnectedThread r;
			// Synchronize a copy of the ConnectedThread
            lock (this)
            {
                if (_state != STATE_CONNECTED)
                    return;
                r = connectedThread;
            }
			// Perform the write unsynchronized
			r.Write (@out);
		}
        #endregion //Write


        #region Read
        /// <summary>
        /// 追加した物
        /// </summary>
        /// <param name='out'>
        /// The bytes to write.
        /// </param>
        public void Read (out byte[] @in, int length)
        {
            // Create temporary object
            ConnectedThread r;
            // Synchronize a copy of the ConnectedThread
            lock (this)
            {
                if (_state != STATE_CONNECTED)
                {
                    @in = new byte[0];
                    return;
                }
                r = connectedThread;
            }

            // Perform the read unsynchronized
            r.Read (out @in, length);
        }
        #endregion //Read



	
        #region ConnectionFailed
		/// <summary>
		/// Indicate that the connection attempt failed and notify the UI Activity.
		/// </summary>
		private void ConnectionFailed ()
		{
			SetState (STATE_LISTEN);
			
			// Send a failure message back to the Activity
			var msg = _handler.ObtainMessage (MESSAGE_TOAST);
			Bundle bundle = new Bundle ();
            //bundle.PutString (MainActivity.TOAST, "Unable to connect device");
            bundle.PutString (MainActivity.TOAST, _context.GetString(Resource.String.not_connected));
			msg.Data = bundle;
			_handler.SendMessage (msg);
		}
        #endregion //ConnectionFailed

	
        #region ConnectionLost
		/// <summary>
		/// Indicate that the connection was lost and notify the UI Activity.
		/// </summary>
		public void ConnectionLost ()
		{
			SetState (STATE_LISTEN);
			
			// Send a failure message back to the Activity
            var msg = _handler.ObtainMessage (MESSAGE_TOAST);
			Bundle bundle = new Bundle ();
			//bundle.PutString (MainActivity.TOAST, "Device connection was lost");
            bundle.PutString (MainActivity.TOAST, _context.GetString(Resource.String.connection_was_lost));
			msg.Data = bundle;
			_handler.SendMessage (msg);
		}
        #endregion //ConnectionLost


        #region AcceptThreadクラス
		/// <summary>
		/// This thread runs while listening for incoming connections. It behaves
		/// like a server-side client. It runs until a connection is accepted
		/// (or until cancelled).
        /// 受信接続を待機している間、このスレッドが実行されます。 
        /// これは、サーバー側のクライアントのように動作します。 
        /// 接続が受け入れられる（またはキャンセルされるまで）されるまで、それは実行されます。
		/// </summary>
		// TODO: Convert to a .NET thread
		private class AcceptThread : Thread
		{
			// The local server socket
			private BluetoothServerSocket mmServerSocket;
            private BluetoothService _service;
			
            #region コンストラクタ
            /// <summary>
            /// Initializes a new instance of the <see cref="BluetoothCom.BluetoothComService+AcceptThread"/> class.
            /// </summary>
            /// <param name="service">Service.</param>
            public AcceptThread (BluetoothService service)
			{
				_service = service;
				BluetoothServerSocket tmp = null;
	
				// Create a new listening server socket
				try
                {
					tmp = _service._adapter.ListenUsingRfcommWithServiceRecord (NAME, MY_UUID);	
				}
                catch (Java.IO.IOException e)
                {
					Log.Error (TAG, "listen() failed", e);
                }
				mmServerSocket = tmp;
			}
            #endregion //コンストラクタ
	

            #region Run
			public override void Run ()
			{
				if (Debug) Log.Debug (TAG, "BEGIN mAcceptThread " + this.ToString ());
				
				Name = "AcceptThread";
				BluetoothSocket socket = null;
	
				// Listen to the server socket if we're not connected
                while (_service._state != BluetoothService.STATE_CONNECTED)
                {
                    try
                    {
                        // This is a blocking call and will only return on a
                        // successful connection or an exception
                        socket = mmServerSocket.Accept();
                    }
                    catch (Java.IO.IOException e)
                    {
                        Log.Error(TAG, "accept() failed", e);
                        break;
                    }
					
                    // If a connection was accepted
                    if (socket != null)
                    {
                        lock (this)
                        {
                            switch (_service._state)
                            {
                                case STATE_LISTEN:
                                case STATE_CONNECTING:
								// Situation normal. Start the connected thread.
                                    _service.Connected(socket, socket.RemoteDevice);
                                    break;
                                case STATE_NONE:
                                case STATE_CONNECTED:
								// Either not ready or already connected. Terminate new socket.
                                    try
                                    {
                                        socket.Close();
                                    }
                                    catch (Java.IO.IOException e)
                                    {
                                        Log.Error(TAG, "Could not close unwanted socket", e);
                                    }
                                    break;
                            }
                        }
                    }
                }
				
				if (Debug) Log.Info (TAG, "END mAcceptThread");
			}
            #endregion //Run
	

            #region Cancel
			public void Cancel ()
			{
				if (Debug) Log.Debug (TAG, "cancel " + this.ToString ());
				
				try
                {
					mmServerSocket.Close ();
				}
                catch (Java.IO.IOException e)
                {
                    Log.Error(TAG, "close() of server failed", e);
                }
			}
            #endregion //Cancel
		}
        #endregion //AcceptThreadクラス


        #region ConnectThreadクラス
		/// <summary>
		/// This thread runs while attempting to make an outgoing connection
		/// with a device. It runs straight through; the connection either
		/// succeeds or fails.
		/// </summary>
		// TODO: Convert to a .NET thread
		protected class ConnectThread : Thread
		{
			private BluetoothSocket mmSocket;
			private BluetoothDevice mmDevice;
            private BluetoothService _service;
			
            #region コンストラクタ
            /// <summary>
            /// Initializes a new instance of the <see cref="BluetoothCom.BluetoothComService+ConnectThread"/> class.
            /// </summary>
            /// <param name="device">Device.</param>
            /// <param name="service">Service.</param>
            public ConnectThread (BluetoothDevice device, BluetoothService service)
			{
				mmDevice = device;
				_service = service;
				BluetoothSocket tmp = null;
				
				// Get a BluetoothSocket for a connection with the
				// given BluetoothDevice
				try 
                {
					tmp = device.CreateRfcommSocketToServiceRecord (MY_UUID);
				} 
                catch (Java.IO.IOException e) 
                {
					Log.Error (TAG, "create() failed", e);
				}
				mmSocket = tmp;
			}
            #endregion //コンストラクタ


            #region Run
			public override void Run ()
			{
				Log.Info (TAG, "BEGIN mConnectThread");
				Name = "ConnectThread";
	
				// Always cancel discovery because it will slow down a connection
				_service._adapter.CancelDiscovery ();
	
				// Make a connection to the BluetoothSocket
				try 
                {
					// This is a blocking call and will only return on a
					// successful connection or an exception
					mmSocket.Connect ();
				} 
                catch //(Java.IO.IOException e)
                {
					_service.ConnectionFailed ();
					// Close the socket
					try 
                    {
						mmSocket.Close ();
					} 
                    catch (Java.IO.IOException e2) 
                    {
						Log.Error (TAG, "unable to close() socket during connection failure", e2);
					}

					// Start the service over to restart listening mode
					_service.Start ();
					return;
				}
	
				// Reset the ConnectThread because we're done
				lock (this) 
                {
					_service.connectThread = null;
				}
	
				// Start the connected thread
				_service.Connected (mmSocket, mmDevice);
			}
            #endregion //Run
	

            #region Cancel
			public void Cancel ()
			{
				try 
                {
					mmSocket.Close ();
				}
                catch (Java.IO.IOException e) 
                {
					Log.Error (TAG, "close() of connect socket failed", e);
				}
			}
            #endregion //Cancel
		}
        #endregion //ConnectThreadクラス

	
        #region ConnectedThreadクラス
		/// <summary>
		/// This thread runs during a connection with a remote device.
		/// It handles all incoming and outgoing transmissions.
        /// このスレッドは、リモートデバイスとの接続時に実行されます。 
        /// それは、すべての受信および送信の送信を処理します。
		/// </summary>
		// TODO: Convert to a .NET thread
		private class ConnectedThread : Thread
		{
            private BluetoothSocket mmSocket;
			private Stream mmInStream;
			private Stream mmOutStream;
            private BluetoothService _service;
	


            #region コンストラクタ
            /// <summary>
            /// Initializes a new instance of the <see cref="BluetoothCom.BluetoothComService+ConnectedThread"/> class.
            /// </summary>
            /// <param name="socket">Socket.</param>
            /// <param name="service">Service.</param>
            public ConnectedThread (BluetoothSocket socket, BluetoothService service)
			{
				Log.Debug (TAG, "create ConnectedThread: ");
				mmSocket = socket;
				_service = service;
				Stream tmpIn = null;
				Stream tmpOut = null;

	
				// Get the BluetoothSocket input and output streams
				try
                {
					tmpIn = socket.InputStream;
					tmpOut = socket.OutputStream;
				}
                catch (Java.IO.IOException e) 
                {
					Log.Error (TAG, "temp sockets not created", e);
				}
                catch
                {
                    throw;
                }
	
				mmInStream = tmpIn;
				mmOutStream = tmpOut;
			}
            #endregion //コンストラクタ
	

            #region Run
            /// <summary>
            /// Starts executing the active part of the class' code.
            /// クラスのコードのアクティブな部分の実行を開始します。
            /// (Run()には、別スレッドで実行する物を書くみたい)
            /// </summary>
			public override void Run ()
			{
				Log.Info (TAG, "BEGIN mConnectedThread");
			}
            #endregion //Run
	

            #region Write
			/// <summary>
			/// Write to the connected OutStream.
			/// </summary>
			/// <param name='buffer'>
			/// The bytes to write
			/// </param>
			public void Write (byte[] buffer)
			{
				try
                {
					mmOutStream.Write (buffer, 0, buffer.Length);
	
					// Share the sent message back to the UI Activity
					_service._handler.ObtainMessage (MESSAGE_WRITE, -1, -1, buffer).SendToTarget ();
				}
                catch (Java.IO.IOException e) 
                {
                    Log.Error(TAG, "Exception during write", e);
                    throw;
                }
			}
            #endregion //Write
	

            #region Read
            /// <summary>
            /// 追加したReadメソッド
            /// </summary>
            /// <param name="buffer">Buffer.</param>
            public void Read (out byte[] buffer, int length)
            {
                buffer = new byte[length];

                try
                {
                    int ret;
                    int j;
                    int timeout = TIMEOUT_READ / TIMEOUT_READ_BYTE;
                    for(int i=0; i<length; i++)
                    {
                        // 自前でタイムアウト処理
                        j = 0;
                        while(true)
                        {
                            if(mmInStream.IsDataAvailable())
                            {
                                ret = mmInStream.Read(buffer, i, 1);    // memo: 読み取るデータがないと、そこで止まる？？？キャンセルはできている
                                if(ret > 0)
                                {
                                    j = 0;      // タイムアウトカントをクリアする
                                    break;
                                }
                            }
                            
                            if(j > timeout)   // 200ms x 10 = 2秒のタイムアウト
                            {
                                throw new Exception(string.Format("Readタイムアウト。データが足りません。\n→要求{0}byte/バッファ{1}byte", length, i));
                            }

                            j++;
                            Sleep(TIMEOUT_READ_BYTE);
                        }
                    }

                    // Share the sent message back to the UI Activity
                    _service._handler.ObtainMessage (MESSAGE_READ, length, -1, buffer).SendToTarget ();
                }
                catch //(Java.IO.IOException e) 
                {
                    //Log.Error(TAG, "Exception during read", e);
                    throw;
                }
            }
            #endregion //Read


            #region Cancel
			public void Cancel ()
			{
				try 
                {
					mmSocket.Close ();
				} 
                catch (Java.IO.IOException e) 
                {
					Log.Error (TAG, "close() of connect socket failed", e);
				}
			}
            #endregion //Cancel
		}
        #endregion //ConnectedThreadクラス
	}
}

